from collections.abc import Generator
from datetime import datetime
from functools import cached_property
from typing import Optional
from urllib.parse import urljoin

import requests
from pydantic import BaseModel, Field, PrivateAttr

from pysilpo.authorization import User
from pysilpo.enums import PayTypeEnum
from pysilpo.utils import get_logger, subtract_months


class ChequeRewardModel(BaseModel):
    reward_type_id: int = Field(..., alias="rewardTypeId")
    apply_text: str = Field(..., alias="applyText")
    reward_value: float = Field(..., alias="rewardValue")
    sign_text: str = Field(..., alias="signText")
    unit_text: str = Field(..., alias="unitText")
    gold_coupon_promo_type: Optional[str] = Field(..., alias="goldCouponPromoType")
    promo_id: Optional[int] = Field(..., alias="promoId")
    used_text: Optional[str] = Field(..., alias="usedText")
    used_count_text: Optional[str] = Field(..., alias="usedCountText")
    used_unit_text: Optional[str] = Field(..., alias="usedUnitText")

    class Meta:
        alternative_name = "chequeRewards"


class ChequePositionModel(BaseModel):
    cheque_line_id: int = Field(..., alias="chequeLineId")
    lager_id: int = Field(..., alias="lagerId")
    lager_name_ua: str = Field(..., alias="lagerNameUA")
    lager_unit: str = Field(..., alias="lagerUnit")
    count: float = Field(..., alias="kolvo")
    price_out: float = Field(..., alias="priceOut")
    unit_text: str = Field(..., alias="unitText")
    image_url: Optional[str] = Field(None, alias="fileName")
    sum_cashback_line: float = Field(default=0.0, alias="sumCashbackLine")

    def get_full_image_url(self, width: int = 480, height: int = 480) -> Optional[str]:
        if self.image_url is not None:
            size = f"{width}x{height}wwm"
            return f"https://content.silpo.ua/sku/ecommerce/{int(self.lager_id / 1e4)}/{size}/{self.lager_id}_{size}_{self.image_url}"
        return None

    class Meta:
        alternative_name = "chequeLines"


class ChequeActionModel(BaseModel):
    action_type: int = Field(..., alias="actionType")  # TODO: Create an enum
    action_type_code_name: Optional[str] = Field(None, alias="actionTypeCodeName")
    action_id: int = Field(..., alias="actionId")
    discpercent: Optional[float] = Field(default=None, alias="discPercent")
    discount: Optional[float] = Field(None, alias="discount")
    varchar_data: Optional[str] = Field(None, alias="varcharData")


class ChequeDetailModel(BaseModel):
    cheque_header: "ChequeModel" = Field(..., alias="chequeHeader")
    sum_discount: float = Field(..., alias="sumDiscount")
    positions: Optional[list[ChequePositionModel]] = Field(..., alias="chequeLines")
    actions: Optional[list[ChequeActionModel]] = Field(..., alias="chequeActions")
    ch_prediction: str = Field(..., alias="chPrediction")
    sum_cashback: float = Field(..., alias="sumCashback")


class ChequeModel(BaseModel):
    _cheque_service: Optional["Cheque"] = PrivateAttr(default=None)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._cheque_service = kwargs.pop("cheque_service", None)

    loyalty_fact_id: int = Field(..., alias="loyaltyFactId")
    sum_reg: float = Field(..., alias="sumReg")
    sum_balance: float = Field(..., alias="sumBalance")
    filial_name: str = Field(..., alias="filialName")
    city_name: str = Field(..., alias="cityName")
    fr_id: int = Field(..., alias="frId")
    z_id: int = Field(..., alias="zId")
    fr_cheque_id: int = Field(..., alias="frChequeId")
    pay_type: PayTypeEnum = Field(..., alias="payType")
    filial_id: int = Field(..., alias="filId")
    cheque_id: int = Field(..., alias="chequeId")
    created: datetime
    fiscal_number: str = Field(..., alias="fiscalNumber")
    business_card_id: int = Field(..., alias="businessCardId")
    external_operation_id: str = Field(..., alias="externalOperationId")

    @cached_property
    def detail(self) -> ChequeDetailModel:
        if self._cheque_service is None:
            raise ValueError("Cheque service is not set")
        return self._cheque_service.get_detail(
            self.cheque_id,
            self.created,
            self.filial_id,
            self.loyalty_fact_id,
        )


class Cheque:
    logger = get_logger("pysilpo.cheque.Cheque")
    _DOMAIN = "https://loyalty-platform-public-api.silpo.ua"
    _ALL_CHEQUES_URL = urljoin(_DOMAIN, "/api/v1/profile/my/cheque/cheque-headers")
    _CHEQUE_DETAIL_URL = urljoin(_DOMAIN, "/api/v1/profile/my/cheque/cheque-info")

    def __init__(self, user: User):
        self.user = user

    def get_detail(
        self, cheque_id: int, created: datetime, fill_id: int, loyalty_fact_id: int
    ) -> ChequeDetailModel:
        payload = {
            "filId": fill_id,
            "chequeId": cheque_id,
            "created": created.isoformat(),
            "loyaltyFactId": loyalty_fact_id,
        }
        self.logger.debug(f"Fetching cheque detail for {payload}")
        resp = requests.post(
            self._CHEQUE_DETAIL_URL,
            json=payload,
            headers={"Authorization": f"Bearer {self.user.access_token}"},
        )
        resp.raise_for_status()
        return ChequeDetailModel(**resp.json())

    def get_all(
        self,
        date_from: Optional[datetime] = None,
        date_to: Optional[datetime] = None,
        page_size: int = 0,
        row_number: int = 0,
    ) -> Generator[ChequeModel, None, None]:
        if date_to is None:
            date_to = datetime.now()

        # Start with the latest 3-month chunk and work backwards
        current_date_to = date_to
        current_date_from = max(
            subtract_months(current_date_to, 3),
            date_from if date_from is not None else datetime.min,
        )

        first_cheque_id_in_chunk: Optional[int] = None

        while date_from is None or current_date_to > date_from:
            payload = {
                "rowNumber": row_number,
                "pageSize": page_size,
                "dateStart": current_date_from.isoformat(),
                "dateEnd": current_date_to.isoformat(),
            }
            self.logger.debug(f"Fetching cheques from {current_date_from} to {current_date_to}")
            resp = requests.post(
                self._ALL_CHEQUES_URL,
                json=payload,
                headers={"Authorization": f"Bearer {self.user.access_token}"},
            )
            resp.raise_for_status()
            data = resp.json()

            # If no more data, exit the loop, but it's unlikely to happen, see the next check
            if not data:
                break

            """
            Silpo API returns data even if it's out of the date range, so we need to prevent duplicates by checking
            if the first cheque ID in the chunk is the same as the previous one.
            """
            if data[0]["chequeId"] == first_cheque_id_in_chunk:
                break  # No more data
            first_cheque_id_in_chunk = data[0]["chequeId"]

            # Yield each item in a flat structure
            for item in data:
                yield ChequeModel(**item, cheque_service=self)

            # Update dates for the next iteration (previous 3-month chunk)
            current_date_to = subtract_months(current_date_from, 3)
            current_date_from = max(
                subtract_months(current_date_to, 3),
                subtract_months(datetime.fromisoformat(data[-1]["created"]), 3)
                if date_from is None
                else date_from,
            )



import logging
import os
from pysilpo.authorization import User
from pysilpo.cheque import Cheque

__version__ = "0.2.1"

DEBUG = int(os.getenv("DEBUG", "0"))

logging.basicConfig(
    level=logging.DEBUG if DEBUG else logging.INFO,
    format="%(asctime)s | %(levelname)-17s | [%(name)s] | %(message)s",
    datefmt="%Y-%m-%d %H:%M:%S",
)

__all__ = ('__version__', 'User', 'Cheque')


import logging
from datetime import datetime, timedelta

import jwt


def get_logger(name: str = "PySilpo") -> logging.Logger:
    return logging.getLogger(name)


def get_jwt_expires_in(jwt_token: str) -> int:
    decoded = jwt.decode(jwt_token, options={"verify_signature": False})
    return decoded["exp"]


def subtract_months(date: datetime, months: int) -> datetime:
    # Calculate target year and month
    target_month = date.month - months
    target_year = date.year + (target_month - 1) // 12  # Adjust year if month < 1
    target_month = (target_month - 1) % 12 + 1  # Adjust month to range 1-12

    # Calculate last day of the target month
    last_day_of_month = (datetime(target_year, target_month + 1, 1) - timedelta(days=1)).day
    target_day = min(date.day, last_day_of_month)  # Ensure valid day of month

    # Return new date with adjusted year, month, and day
    return datetime(target_year, target_month, target_day, date.hour, date.minute, date.second)



import base64
import hashlib
import random
import re
import secrets
import sys
import time
from datetime import datetime

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc

from functools import cached_property
from typing import Literal, Optional
from urllib.parse import parse_qs, urljoin, urlparse

import requests
from pydantic import BaseModel, model_validator

from pysilpo.cache import SQLiteCache
from pysilpo.exceptions import SilpoAuthorizationException, SilpoOTPInvalidException
from pysilpo.utils import get_jwt_expires_in, get_logger


class Token(BaseModel):
    id_token: str
    access_token: str
    expires_in: datetime
    token_type: str
    scope: str

    @model_validator(mode="before")
    @classmethod
    def set_expire_time(cls, values):
        values["expires_in"] = get_jwt_expires_in(values["access_token"])
        return values


class User:
    """
    All public methods should return self to allow chaining.
    """

    logger = get_logger("pysilpo.authorization.User")
    _base_auth_domain = "https://auth.silpo.ua"
    _openid_configuration = urljoin(_base_auth_domain, "/.well-known/openid-configuration")
    _phone_number_pattern = r"^\+380\d{9}$"

    def __init__(
        self,
        phone_number: str,
        otp_delivery_method: Literal["sms", "viber-sms"] = "sms",
        openid_client_id: Optional[str] = "profile--profile--cabinet",
        openid_scope: Optional[
            str
        ] = "openid public-my profile--security--identity-service:internal-api--call "
        "core--core--media-service:media--upload payments--payments--wallet-service:cards--read-my "
        "core--core--media-service:media--upload",
        openid_redirect_uri: Optional[str] = "https://id.silpo.ua/signin-oidc",
    ):
        if not re.match(self._phone_number_pattern, phone_number):
            raise ValueError("Invalid phone number, must be in format +380XXYYYYYYY")
        self.phone_number = phone_number
        self.session = requests.Session()
        self.otp_delivery_method = otp_delivery_method
        self.client_id = openid_client_id
        self.scope = openid_scope
        self.redirect_uri = openid_redirect_uri
        self.code_verifier = secrets.token_urlsafe(64)

        self.token: Optional[Token] = self.cached_token

    @cached_property
    def openid_configuration(self) -> dict:
        resp = self.session.get(self._openid_configuration)
        resp.raise_for_status()
        return resp.json()

    @property
    def cached_token(self) -> Optional[Token]:
        return SQLiteCache().get(f"token_{self.phone_number}")

    def request_otp(
        self, delivery_method: Optional[Literal["sms", "viber-sms"]] = None, force: bool = False
    ) -> "User":
        if self.token and not force:
            self.logger.debug(
                "[request_otp] Already logged in with token scope. Skipping OTP request"
            )
            return self

        auth_cookies = SQLiteCache().get(f"cookie_{self.phone_number}")
        if auth_cookies and not force:
            self.logger.debug("[request_otp] Have cookies for authorization. Skipping OTP request")
            return self

        if delivery_method is None:
            delivery_method = self.otp_delivery_method
        full_url = urljoin(self._base_auth_domain, "/api/v2/Login/ByPhone")
        json = {
            "phone": self.phone_number,
            "recaptcha": None,
            "delivery_method": delivery_method,
            "phoneChannelType": 0,
        }
        self.logger.debug(f"[_request_otp] Requesting OTP with {json} to {full_url}")
        resp = self.session.post(full_url, json=json)
        json_data = resp.json()
        self.logger.debug(f"[_request_otp] Received response: {json_data}")
        if not resp.ok:
            if json_data["secondsTillNextOTP"]:
                self.logger.warning(
                    f"[_request_otp] Got error from Silpo, need to wait till next OTP: {json_data['secondsTillNextOTP']} seconds. Waiting..."
                )
                time.sleep(json_data["secondsTillNextOTP"] + random.randint(1, 4))
                return self.request_otp(delivery_method)
            raise SilpoAuthorizationException(f"Error while requesting OTP: {json_data}")
        return self

    def _verify_otp(self, otp_code: str) -> "User":
        """
        This method send OTP code to Silpo API to verify it.
        If OTP code is valid, Silpo API will set cookies required for the next requests, otherwise it will raise an exception.

        :param otp_code: 6-digit OTP code
        :return: None
        """
        if not re.match(r"^\d{6}$", otp_code):
            raise SilpoOTPInvalidException("OTP code must contain 6 digits")
        full_url = urljoin(self._base_auth_domain, "/api/v2/Login/LoginWithOTP")
        json = {
            "phone": self.phone_number,
            "otp": otp_code,
            "phoneChannelType": 0,
        }
        self.logger.debug(f"[_verify_otp] Verifying OTP with {json} to {full_url}")
        resp = self.session.post(full_url, json=json)
        json_data = resp.json()
        self.logger.debug(
            f"[_verify_otp] Received response: {json_data}. With cookies: {resp.cookies}"
        )
        if not resp.ok or json_data["error"]:
            raise SilpoOTPInvalidException(f"Error while verifying OTP: {json_data}")
        SQLiteCache().set(f"cookie_{self.phone_number}", dict(resp.cookies))
        return self

    @staticmethod
    def _enter_cli_otp() -> str:
        while True:
            otp_code = input("Enter OTP code: ")
            if not re.match(r"^\d{6}$", otp_code):
                raise SilpoOTPInvalidException("OTP code must contain 6 digits")
            break
        return otp_code

    def _openid_authorize(self, auth_cookies: Optional[dict] = None) -> str:
        """
        This method authorizes user with OpenID Connect protocol.
        It sends a request to Silpo API to get authorization code.
        When request is finished, it parses the response.url query params to get authorization code.
        This code will be used to get access token.
        :return: OpenID authorization code
        """
        full_url = self.openid_configuration["authorization_endpoint"]
        code_challenge = (
            base64.urlsafe_b64encode(hashlib.sha256(self.code_verifier.encode()).digest())
            .rstrip(b"=")
            .decode("ascii")
        )
        params = {
            "client_id": self.client_id,
            "redirect_uri": self.redirect_uri,
            "response_type": "code",
            "scope": self.scope,
            "state": secrets.token_urlsafe(16),
            "code_challenge": code_challenge,
            "code_challenge_method": "S256",
            "response_mode": "query",
        }
        self.logger.debug(f"[_openid_authorize] Authorizing with {params} to {full_url}")
        resp = self.session.get(full_url, params=params, cookies=auth_cookies)
        resp.raise_for_status()
        self.logger.debug(
            f"[_openid_authorize] Received location: {resp.url}. With headers: {resp.headers} and cookies: {resp.cookies}"
        )

        parsed_url = urlparse(resp.url)
        query_params = parse_qs(parsed_url.query)
        self.logger.debug(f"[_openid_authorize] Parsed query params: {query_params}")
        auth_code = query_params.get("code", [None])[0]

        if auth_code is None:
            raise SilpoAuthorizationException("No auth code in response for OpenID authorization")
        return auth_code

    def _get_token(self, auth_code: str) -> Token:
        full_url = self.openid_configuration["token_endpoint"]
        form_data = {
            "client_id": self.client_id,
            "code": auth_code,
            "redirect_uri": self.redirect_uri,
            "code_verifier": self.code_verifier,
            "grant_type": "authorization_code",
        }
        self.logger.debug(
            f"[_get_access_token] Getting access token with {form_data} to {full_url}"
        )
        resp = self.session.post(full_url, data=form_data)
        json_data = resp.json()
        self.logger.debug(f"[_get_access_token] Received response: {json_data}")
        if not resp.ok:
            raise SilpoAuthorizationException(f"Error while getting access token: {json_data}")
        return Token(**json_data)

    def is_expired(self) -> bool:
        if self.cached_token is None:
            return True
        return self.token.expires_in < datetime.now(tz=UTC)

    def _refresh_token(self) -> None:
        auth_cookies = SQLiteCache().get(f"cookie_{self.phone_number}")
        if not auth_cookies:
            raise SilpoAuthorizationException(
                "No cookies found for token refresh."
                "Please login first using User(phone_number=...).request_otp().login() method."
            )
        code = self._openid_authorize(auth_cookies)
        self.token = self._get_token(code)
        self.logger.debug(
            f"[refresh_token] Token refreshed with scope: {self.token.scope} | {self.token.expires_in} UTC"
        )

    @property
    def access_token(self) -> str:
        if self.token is None:
            raise SilpoAuthorizationException(
                "Access token is not set. "
                "Please login first using User(phone_number=...).request_otp().login() method."
            )
        if self.is_expired():
            self._refresh_token()
        return self.token.access_token

    def set_token(self, token: Token) -> "User":
        self.token = token
        self.logger.debug(
            f"[set_token] Set token scope: {self.token.scope} | {self.token.expires_in} UTC"
        )
        SQLiteCache().set(f"token_{self.phone_number}", token, expires_in=token.expires_in)
        return self

    def login(self, otp_code: Optional[str] = None, force: bool = False) -> "User":
        if self.token and not force:
            self.logger.debug(
                f"[login] Already logged in with token scope: {self.token.scope} | {self.token.expires_in} UTC"
            )
            return self

        auth_cookies = SQLiteCache().get(f"cookie_{self.phone_number}")
        if not auth_cookies and not force:
            if otp_code is None:
                otp_code = self._enter_cli_otp()
            self._verify_otp(otp_code)

        auth_code = self._openid_authorize(auth_cookies)
        self.token = self._get_token(auth_code)
        self.logger.debug(
            f"[login] Logged in with token scope: {self.token.scope} | {self.token.expires_in} UTC"
        )
        SQLiteCache().set(
            f"token_{self.phone_number}", self.token, expires_in=self.token.expires_in
        )
        return self



class SilpoException(Exception):
    pass


class SilpoRequestException(SilpoException):
    pass


class SilpoAuthorizationException(SilpoRequestException):
    pass


class SilpoOTPInvalidException(SilpoAuthorizationException):
    pass



import pickle
import sqlite3
import sys
from datetime import datetime

if sys.version_info >= (3, 11):
    from datetime import UTC
else:
    from datetime import timezone

    UTC = timezone.utc

from pathlib import Path
from typing import Union

MAX_TS = round(datetime.max.replace(year=9998).timestamp())  # Maximum Unix timestamp


class SQLiteCache:
    def __init__(self, db_name="cache.db", use_pickle=True):
        """Initialize with optional pickle support and a database name."""

        user_data_dir = Path.home() / ".pysilpo"
        user_data_dir.mkdir(parents=True, exist_ok=True)  # Ensure the directory exists
        self.db_name = user_data_dir / db_name

        self.conn = sqlite3.connect(str(self.db_name))
        self.cursor = self.conn.cursor()

        # Create the cache table with columns for key, value, and expiry time
        self.cursor.execute(
            """CREATE TABLE IF NOT EXISTS cache (
                                key TEXT PRIMARY KEY,
                                value BLOB,  -- Store value as a binary blob
                                expiry REAL)"""
        )
        self.conn.commit()

        self.use_pickle = use_pickle  # Whether to use pickle for serialization

    def __del__(self):
        """Automatically close the SQLite connection when the object is deleted."""

        if self.conn:
            self.conn.close()

    def _check_expiry(self, key):
        """Check if the cache entry has expired."""
        self.cursor.execute("SELECT expiry FROM cache WHERE key = ?", (key,))
        result = self.cursor.fetchone()
        if result:
            expiry_time = result[0]
            now = datetime.now(tz=UTC).timestamp()
            return expiry_time < now  # Return True if expired
        return False

    def get(self, key):
        """Retrieve a cached value, or None if expired or not found."""
        # Check if the key exists and if it's expired
        if self._check_expiry(key):
            self.remove(key)  # Remove expired entry
            return None  # Expired

        self.cursor.execute("SELECT value FROM cache WHERE key = ?", (key,))
        result = self.cursor.fetchone()
        if result:
            value = result[0]
            if self.use_pickle:
                return pickle.loads(value)  # Deserialize using pickle
            return value
        return None  # Not found

    def set(self, key, value, expires_in: Union[datetime, int, None] = None):
        """Store a value in the cache with an optional TTL."""
        if expires_in is None:
            expiry_time = MAX_TS
        elif isinstance(expires_in, int):
            expiry_time = expires_in
        else:
            expiry_time = round(expires_in.timestamp())  # Convert datetime to Unix timestamp
        # Serialize the value if using pickle
        if self.use_pickle:
            value = pickle.dumps(value)  # Serialize using pickle

        self.cursor.execute(
            "REPLACE INTO cache (key, value, expiry) VALUES (?, ?, ?)",
            (key, value, expiry_time),
        )
        self.conn.commit()

    def remove(self, key):
        """Remove a key-value pair from the cache."""
        self.cursor.execute("DELETE FROM cache WHERE key = ?", (key,))
        self.conn.commit()

    def clear(self):
        """Clear the entire cache."""
        self.cursor.execute("DELETE FROM cache")
        self.conn.commit()

    def exists(self, key):
        """Check if a key exists in the cache (even if expired)."""
        self.cursor.execute("SELECT 1 FROM cache WHERE key = ?", (key,))
        return self.cursor.fetchone() is not None

    def get_all(self):
        """Retrieve all keys and values from the cache."""
        # TODO: Fix ttl check
        self.cursor.execute("SELECT key, value FROM cache")
        rows = self.cursor.fetchall()
        # If using pickle, deserialize each value
        if self.use_pickle:
            return [(key, pickle.loads(value)) for key, value in rows]
        return rows

    def close(self):
        """Close the SQLite database connection."""
        self.conn.close()
